import pytest

from patreon.schemas import campaign


@pytest.fixture
def attributes():
    return [
        'summary',
        'creation_name',
        'pay_per_name',
        'one_liner',
        'main_video_embed',
        'main_video_url',
        'image_small_url',
        'image_url',
        'thanks_video_url',
        'thanks_embed',
        'thanks_msg',
        'is_monthly',
        'is_nsfw',
        'is_charged_immediately',
        'is_plural',
        'created_at',
        'published_at',
        'pledge_url',
        'pledge_sum',
        'patron_count',
        'creation_count',
        'outstanding_payment_amount_cents',
    ]


@pytest.fixture
def relationships():
    return [
        'rewards',
        'creator',
        'goals',
        'pledges',
        'current_user_pledge',
        'post_aggregation',
        'categories',
        'preview_token',
    ]


def test_schema_attributes_are_properly_formatted(attributes):
    for attribute_name in attributes:
        value = getattr(campaign.Attributes, attribute_name, None)
        assert value is not None and value is attribute_name


def test_schema_relationships_are_properly_formatted(relationships):
    for relationship_name in relationships:
        value = getattr(campaign.Relationships, relationship_name, None)
        assert value is not None and value is relationship_name


def test_schema_has_expected_default_attributes():
    assert campaign.Attributes.summary in campaign.default_attributes
    assert campaign.Attributes.creation_name in campaign.default_attributes
    assert campaign.Attributes.pay_per_name in campaign.default_attributes
    assert campaign.Attributes.one_liner in campaign.default_attributes
    assert campaign.Attributes.main_video_embed in campaign.default_attributes
    assert campaign.Attributes.main_video_url in campaign.default_attributes
    assert campaign.Attributes.image_small_url in campaign.default_attributes
    assert campaign.Attributes.image_url in campaign.default_attributes
    assert campaign.Attributes.thanks_video_url in campaign.default_attributes
    assert campaign.Attributes.thanks_embed in campaign.default_attributes
    assert campaign.Attributes.thanks_msg in campaign.default_attributes
    assert campaign.Attributes.is_monthly in campaign.default_attributes
    assert campaign.Attributes.is_nsfw in campaign.default_attributes
    assert campaign.Attributes.is_charged_immediately in campaign.default_attributes
    assert campaign.Attributes.is_plural in campaign.default_attributes
    assert campaign.Attributes.created_at in campaign.default_attributes
    assert campaign.Attributes.published_at in campaign.default_attributes
    assert campaign.Attributes.pledge_url in campaign.default_attributes
    assert campaign.Attributes.pledge_sum in campaign.default_attributes
    assert campaign.Attributes.patron_count in campaign.default_attributes
    assert campaign.Attributes.creation_count in campaign.default_attributes
    assert campaign.Attributes.outstanding_payment_amount_cents in campaign.default_attributes


def test_schema_has_expected_default_relationships():
    assert campaign.Relationships.rewards in campaign.default_relationships
    assert campaign.Relationships.creator in campaign.default_relationships
    assert campaign.Relationships.goals in campaign.default_relationships



import mock
import pytest

from patreon import oauth
from patreon.utils import user_agent_string

MOCK_CLIENT_ID = 'Mock Client ID'
MOCK_CLIENT_SECRET = 'Mock Client Secret'
MOCK_CODE = 'Mock Code'
MOCK_REDIRECT_URI = 'https://www.patreon.com/members'
MOCK_REFRESH_TOKEN = 'Mock Refresh Token'


@pytest.fixture
def client():
    return oauth.OAuth(MOCK_CLIENT_ID, MOCK_CLIENT_SECRET)


@mock.patch('requests.post')
def test_get_tokens_requests_tokens_as_expected(post, client):
    client.get_tokens(
        code=MOCK_CODE,
        redirect_uri=MOCK_REDIRECT_URI,
    )

    post.assert_called_once_with(
        'https://www.patreon.com/api/oauth2/token',
        params={
            'grant_type': 'authorization_code',
            'code': MOCK_CODE,
            'client_id': MOCK_CLIENT_ID,
            'client_secret': MOCK_CLIENT_SECRET,
            'redirect_uri': MOCK_REDIRECT_URI,
        },
        headers={
            'User-Agent': user_agent_string(),
        },
    )


@mock.patch('requests.post')
def test_refresh_token_gets_a_new_token_as_expected(post, client):
    client.refresh_token(
        redirect_uri=MOCK_REDIRECT_URI,
        refresh_token=MOCK_REFRESH_TOKEN,
    )

    post.assert_called_once_with(
        'https://www.patreon.com/api/oauth2/token',
        params={
            'grant_type': 'refresh_token',
            'refresh_token': MOCK_REFRESH_TOKEN,
            'client_id': MOCK_CLIENT_ID,
            'client_secret': MOCK_CLIENT_SECRET,
        },
        headers={
            'User-Agent': user_agent_string(),
        },
    )



class Attributes(object):
    amount_cents = 'amount_cents'
    title = 'title'
    description = 'description'
    created_at = 'created_at'
    reached_at = 'reached_at'


class Relationships(object):
    pass


default_attributes = [
    Attributes.amount_cents,
    Attributes.title,
    Attributes.description,
    Attributes.created_at,
    Attributes.reached_at,
]

default_relationships = [
]



import datetime
import functools
import mock

from patreon import api
from patreon.jsonapi import url_util
from patreon.jsonapi.parser import JSONAPIParser
from patreon.utils import user_agent_string
from patreon.version_compatibility import urllib_parse
from patreon.version_compatibility.utc_timezone import utc_timezone

MOCK_CAMPAIGN_ID = 12
API_ROOT_ENDPOINT = 'https://www.patreon.com/api/oauth2/api/'
MOCK_ACCESS_TOKEN = 'mock token'
MOCK_CURSOR_VALUE = 'Mock Cursor Value'

DEFAULT_API_HEADERS = {
    'Authorization': 'Bearer ' + MOCK_ACCESS_TOKEN,
    'User-Agent': user_agent_string(),
}

client = api.API(access_token=MOCK_ACCESS_TOKEN)


def api_url(*segments, **query):
    path = '/'.join(map(str, segments))

    fields = query.get('fields', None)
    includes = query.get('includes', None)

    if fields:
        del query['fields']

    if includes:
        del query['includes']

    if query:
        path += '?' + urllib_parse.urlencode(query)

    return url_util.build_url(
        API_ROOT_ENDPOINT + path,
        fields=fields,
        includes=includes,
    )


def assert_valid_api_call(method, api_url, query=None, **kwargs):
    kwargs.setdefault('headers', DEFAULT_API_HEADERS)
    method.assert_called_once_with(api_url, **kwargs)


class MockResponse(object):
    def __init__(self, data=None, status_code=200):
        self.data = data or {}

        self.status_code = status_code

    def json(self):
        return self.data


def api_test(method='GET', **response_kwargs):
    """ Decorator to ensure API calls are made and return expected data. """

    method = method.lower()

    def api_test_factory(fn):
        @functools.wraps(fn)
        @mock.patch('requests.{}'.format(method))
        def execute_test(method_func, *args, **kwargs):
            method_func.return_value = MockResponse(**response_kwargs)

            expected_url, response = fn(*args, **kwargs)

            method_func.assert_called_once()
            assert_valid_api_call(method_func, expected_url)
            assert isinstance(response, JSONAPIParser)
            assert response.json_data is method_func.return_value.data

        return execute_test

    return api_test_factory


def test_extract_cursor_returns_cursor_when_provided():
    assert MOCK_CURSOR_VALUE == api.API.extract_cursor(
        {
            'links':
                {
                    'next':
                        'https://patreon.com/members?page[cursor]=' +
                        MOCK_CURSOR_VALUE,
                },
        }
    )


def test_extract_cursor_returns_None_when_no_cursor_provided():
    assert None is api.API.extract_cursor(
        {
            'links': {
                'next': 'https://patreon.com/members?page[offset]=25',
            },
        }
    )


def test_extract_cursor_returns_None_when_link_is_not_a_string():
    assert None is api.API.extract_cursor({
        'links': {
            'next': None,
        },
    })


def test_extract_cursor_returns_None_when_link_is_malformed():
    caught_exception = False

    try:
        api.API.extract_cursor({
            'links': {
                'next': 12,
            },
        })

    except Exception as e:
        caught_exception = True
        assert e.args[0] == 'Provided cursor path did not result in a link'

    assert caught_exception


@api_test()
def test_can_fetch_user():
    return api_url('current_user'), client.fetch_user()


@api_test()
def test_can_fetch_campaign():
    expected_url = api_url('current_user', 'campaigns')
    response = client.fetch_campaign()
    return expected_url, response


@api_test()
def test_can_fetch_api_and_patrons():
    response = client.fetch_campaign_and_patrons()

    expected_url = api_url(
        'current_user',
        'campaigns',
        includes=['rewards', 'creator', 'goals', 'pledges'],
    )

    return expected_url, response


@api_test()
def test_can_fetch_api_and_patrons_with_custom_includes():
    expected_url = api_url(
        'current_user',
        'campaigns',
        includes=['creator'],
    )

    response = client.fetch_campaign_and_patrons(
        includes=['creator'],
    )

    return expected_url, response


@api_test()
def test_can_fetch_page_of_pledges():
    PAGE_COUNT = 25

    response = client.fetch_page_of_pledges(MOCK_CAMPAIGN_ID, PAGE_COUNT)

    query_params = {'page[count]': PAGE_COUNT}

    expected_url = api_url(
        'campaigns', MOCK_CAMPAIGN_ID, 'pledges', **query_params
    )

    return expected_url, response


@api_test()
def test_can_fetch_page_of_pledges_with_arbitrary_cursor():
    PAGE_COUNT = 25
    MOCK_CURSOR = 'Mock Cursor'

    response = client.fetch_page_of_pledges(
        MOCK_CAMPAIGN_ID,
        PAGE_COUNT,
        cursor=MOCK_CURSOR,
    )

    query_params = {
        'page[count]': PAGE_COUNT,
        'page[cursor]': MOCK_CURSOR,
    }

    expected_url = api_url(
        'campaigns', MOCK_CAMPAIGN_ID, 'pledges', **query_params
    )

    return expected_url, response


@api_test()
def test_can_fetch_page_of_pledges_with_custom_options_without_tzinfo():
    PAGE_COUNT = 25
    MOCK_CURSOR = datetime.datetime.now()
    MOCK_FIELDS = {'field': ['value']}
    MOCK_INCLUDES = ['mock includes']

    EXPECTED_CURSOR = MOCK_CURSOR.replace(tzinfo=utc_timezone()).isoformat()

    response = client.fetch_page_of_pledges(
        MOCK_CAMPAIGN_ID,
        PAGE_COUNT,
        cursor=MOCK_CURSOR,
        includes=MOCK_INCLUDES,
        fields=MOCK_FIELDS,
    )

    query_params = {
        'page[count]': PAGE_COUNT,
        'page[cursor]': EXPECTED_CURSOR,
        'includes': MOCK_INCLUDES,
        'fields': MOCK_FIELDS,
    }

    expected_url = api_url(
        'campaigns', MOCK_CAMPAIGN_ID, 'pledges', **query_params
    )

    return expected_url, response


@api_test()
def test_can_fetch_page_of_pledges_with_custom_options_with_tzinfo():
    PAGE_COUNT = 25
    MOCK_CURSOR = datetime.datetime.now().replace(tzinfo=utc_timezone())
    MOCK_FIELDS = {'field': ['value']}
    MOCK_INCLUDES = ['mock includes']

    EXPECTED_CURSOR = MOCK_CURSOR.isoformat()

    response = client.fetch_page_of_pledges(
        MOCK_CAMPAIGN_ID,
        PAGE_COUNT,
        cursor=MOCK_CURSOR,
        includes=MOCK_INCLUDES,
        fields=MOCK_FIELDS,
    )

    query_params = {
        'page[count]': PAGE_COUNT,
        'page[cursor]': EXPECTED_CURSOR,
        'includes': MOCK_INCLUDES,
        'fields': MOCK_FIELDS,
    }

    expected_url = api_url(
        'campaigns', MOCK_CAMPAIGN_ID, 'pledges', **query_params
    )

    return expected_url, response



import itertools


class JSONAPIParser(object):
    """This is a reader for JSON API Documents"""
    def __init__(self, json_data):
        self.json_data = json_data

    def data(self):
        if isinstance(self.json_data["data"], list):
            return [JSONAPIResource(datum, self) for datum in self.json_data["data"]]
        else:
            return JSONAPIResource(self.json_data["data"], self)

    def find_resource_by_type_and_id(self, resource_type, resource_id):
        for resource_json_data in self._all_resource_json_data():
            if resource_json_data.get("type") == resource_type and resource_json_data.get("id") == resource_id:
                return JSONAPIResource(resource_json_data, self)
        return None

    def _all_resource_json_data(self):
        data = self.json_data["data"]
        if not isinstance(data, list):
            data = [data]
        resources = itertools.chain([], data)

        if "included" in self.json_data:
            resources = itertools.chain(resources, self.json_data["included"])

        return resources


class JSONAPIResource(object):
    """Represents a single resource in a JSON API Document"""
    def __init__(self, json_data, document):
        self.json_data = json_data
        self.document = document

    def id(self):
        return self.json_data.get("id")

    def type(self):
        return self.json_data.get("type")

    def attributes(self):
        return self.json_data.get("attributes", {})

    def attribute(self, name):
        return self.attributes().get(name, None)

    def relationships(self):
        return self.json_data.get("relationships", {})

    def relationship(self, relationship_name):
        relationship_info = self.relationship_info(relationship_name)
        return relationship_info.resource() if relationship_info else None

    def relationship_info(self, relationship_name):
        relationships = self.relationships()
        if relationship_name not in relationships:
            return None

        return JSONAPIRelationshipInfo(relationships[relationship_name], self.document)


class JSONAPIRelationshipInfo(object):
    """Represents a named relationship in a JSON API Document"""
    def __init__(self, json_data, document):
        self.json_data = json_data
        self.document = document

    def resource(self):
        resource_reference = self.resource_reference()

        if isinstance(resource_reference, list):
            return [one_id.resource() for one_id in resource_reference]

        return resource_reference.resource()

    def resource_reference(self):
        if "data" in self.json_data:
            if isinstance(self.json_data["data"], list):
                return [JSONAPIRelatedResource(datum, self.document) for datum in self.json_data["data"]]
            else:
                return JSONAPIRelatedResource(self.json_data["data"], self.document)


class JSONAPIRelatedResource(object):
    """Represents a single resource related to another resource in a JSON API Document"""
    def __init__(self, json_data, document):
        self.json_data = json_data
        self.document = document

    def type(self):
        return self.json_data.get("type")

    def id(self):
        return self.json_data.get("id")

    def resource(self):
        return self.document.find_resource_by_type_and_id(self.type(), self.id())



import pytest

from patreon.schemas import user


@pytest.fixture
def attributes():
    return [
        'email',
        'first_name',
        'last_name',
        'full_name',
        'gender',
        'status',
        'vanity',
        'about',
        'facebook_id',
        'image_url',
        'thumb_url',
        'thumbnails',
        'youtube',
        'twitter',
        'facebook',
        'twitch',
        'is_suspended',
        'is_deleted',
        'is_nuked',
        'created',
        'url',
        'like_count',
        'comment_count',
        'is_creator',
        'hide_pledges',
        'two_factor_enabled',
    ]


@pytest.fixture
def relationships():
    return [
        'pledges',
        'cards',
        'follows',
        'campaign',
        'presence',
        'session',
        'locations',
        'current_user_follow',
        'pledge_to_current_user',
    ]


def test_schema_attributes_are_properly_formatted(attributes):
    for attribute_name in attributes:
        value = getattr(user.Attributes, attribute_name, None)
        assert value is not None and value is attribute_name


def test_schema_relationships_are_properly_formatted(relationships):
    for relationship_name in relationships:
        value = getattr(user.Relationships, relationship_name, None)
        assert value is not None and value is relationship_name


def test_schema_has_expected_default_attributes():
    assert user.Attributes.email in user.default_attributes
    assert user.Attributes.first_name in user.default_attributes
    assert user.Attributes.last_name in user.default_attributes
    assert user.Attributes.full_name in user.default_attributes
    assert user.Attributes.gender in user.default_attributes
    assert user.Attributes.status in user.default_attributes
    assert user.Attributes.vanity in user.default_attributes
    assert user.Attributes.about in user.default_attributes
    assert user.Attributes.facebook_id in user.default_attributes
    assert user.Attributes.image_url in user.default_attributes
    assert user.Attributes.thumb_url in user.default_attributes
    assert user.Attributes.thumbnails in user.default_attributes
    assert user.Attributes.youtube in user.default_attributes
    assert user.Attributes.twitter in user.default_attributes
    assert user.Attributes.facebook in user.default_attributes
    assert user.Attributes.twitch in user.default_attributes
    assert user.Attributes.is_suspended in user.default_attributes
    assert user.Attributes.is_deleted in user.default_attributes
    assert user.Attributes.is_nuked in user.default_attributes
    assert user.Attributes.created in user.default_attributes
    assert user.Attributes.url in user.default_attributes


def test_schema_has_expected_default_relationships():
    assert user.Relationships.campaign in user.default_relationships
    assert user.Relationships.pledges in user.default_relationships



class Attributes(object):
    amount = 'amount'
    amount_cents = 'amount_cents'
    user_limit = 'user_limit'
    remaining = 'remaining'
    description = 'description'
    requires_shipping = 'requires_shipping'
    created_at = 'created_at'
    url = 'url'
    patron_count = 'patron_count'


class Relationships(object):
    creator = 'creator'


default_attributes = [
    Attributes.amount,
    Attributes.amount_cents,
    Attributes.user_limit,
    Attributes.remaining,
    Attributes.description,
    Attributes.requires_shipping,
    Attributes.created_at,
    Attributes.url,
    Attributes.patron_count,
]

default_relationships = [
    Relationships.creator
]



import requests

from patreon.jsonapi.parser import JSONAPIParser
from patreon.jsonapi.url_util import build_url
from patreon.schemas import campaign
from patreon.utils import user_agent_string
from patreon.version_compatibility.utc_timezone import utc_timezone
from patreon.version_compatibility.urllib_parse import urlencode, urlparse, parse_qs


class API(object):
    def __init__(self, access_token):
        super(API, self).__init__()
        self.access_token = access_token

    def fetch_user(self, includes=None, fields=None):
        return self.__get_jsonapi_doc(
            build_url('current_user', includes=includes, fields=fields)
        )

    def fetch_campaign_and_patrons(self, includes=None, fields=None):
        if not includes:
            includes = campaign.default_relationships \
                + [campaign.Relationships.pledges]
        return self.fetch_campaign(includes=includes, fields=fields)

    def fetch_campaign(self, includes=None, fields=None):
        return self.__get_jsonapi_doc(
            build_url(
                'current_user/campaigns', includes=includes, fields=fields
            )
        )

    def fetch_page_of_pledges(
            self, campaign_id, page_size, cursor=None, includes=None,
            fields=None
    ):
        url = 'campaigns/{0}/pledges'.format(campaign_id)
        params = {'page[count]': page_size}
        if cursor:
            try:
                cursor = self.__as_utc(cursor).isoformat()
            except AttributeError:
                pass
            params.update({'page[cursor]': cursor})
        url += "?" + urlencode(params)
        return self.__get_jsonapi_doc(
            build_url(url, includes=includes, fields=fields)
        )

    @staticmethod
    def extract_cursor(jsonapi_document, cursor_path='links.next'):
        def head_and_tail(path):
            if path is None:
                return None, None
            head_tail = path.split('.', 1)
            return head_tail if len(head_tail) == 2 else (head_tail[0], None)

        if isinstance(jsonapi_document, JSONAPIParser):
            jsonapi_document = jsonapi_document.json_data

        head, tail = head_and_tail(cursor_path)
        current_dict = jsonapi_document
        while head and type(current_dict) == dict and head in current_dict:
            current_dict = current_dict[head]
            head, tail = head_and_tail(tail)

        # Path was valid until leaf, at which point nothing was found
        if current_dict is None or (head is not None and tail is None):
            return None
        # Path stopped before leaf was reached
        elif current_dict and type(current_dict) != str:
            raise Exception(
                'Provided cursor path did not result in a link', current_dict
            )

        link = current_dict
        query_string = urlparse(link).query
        parsed_query_string = parse_qs(query_string)
        if 'page[cursor]' in parsed_query_string:
            return parsed_query_string['page[cursor]'][0]
        else:
            return None

    # Internal methods
    def __get_jsonapi_doc(self, suffix):
        response_json = self.__get_json(suffix)
        if response_json.get('errors'):
            return response_json
        return JSONAPIParser(response_json)

    def __get_json(self, suffix):
        response = requests.get(
            "https://www.patreon.com/api/oauth2/api/{}".format(suffix),
            headers={
                'Authorization': "Bearer {}".format(self.access_token),
                'User-Agent': user_agent_string(),
            }
        )
        return response.json()

    @staticmethod
    def __as_utc(dt):
        if hasattr(dt, 'tzinfo'):
            if dt.tzinfo:
                return dt.astimezone(utc_timezone())
            else:
                return dt.replace(tzinfo=utc_timezone())
        return dt



class Attributes(object):
    summary = 'summary'
    creation_name = 'creation_name'
    pay_per_name = 'pay_per_name'
    one_liner = 'one_liner'
    main_video_embed = 'main_video_embed'
    main_video_url = 'main_video_url'
    image_small_url = 'image_small_url'
    image_url = 'image_url'
    thanks_video_url = 'thanks_video_url'
    thanks_embed = 'thanks_embed'
    thanks_msg = 'thanks_msg'
    is_monthly = 'is_monthly'
    is_nsfw = 'is_nsfw'
    is_charged_immediately = 'is_charged_immediately'
    is_charge_upfront_eligible = 'is_charge_upfront_eligible'
    is_plural = 'is_plural'
    created_at = 'created_at'
    published_at = 'published_at'
    pledge_url = 'pledge_url'
    pledge_sum = 'pledge_sum'
    patron_count = 'patron_count'
    creation_count = 'creation_count'
    outstanding_payment_amount_cents = 'outstanding_payment_amount_cents'


class Relationships(object):
    rewards = 'rewards'
    creator = 'creator'
    goals = 'goals'
    pledges = 'pledges'
    current_user_pledge = 'current_user_pledge'
    post_aggregation = 'post_aggregation'
    categories = 'categories'
    preview_token = 'preview_token'


default_attributes = [
    Attributes.summary,
    Attributes.creation_name,
    Attributes.pay_per_name,
    Attributes.one_liner,
    Attributes.main_video_embed,
    Attributes.main_video_url,
    Attributes.image_small_url,
    Attributes.image_url,
    Attributes.thanks_video_url,
    Attributes.thanks_embed,
    Attributes.thanks_msg,
    Attributes.is_monthly,
    Attributes.is_nsfw,
    Attributes.is_charged_immediately,
    Attributes.is_plural,
    Attributes.created_at,
    Attributes.published_at,
    Attributes.pledge_url,
    Attributes.pledge_sum,
    Attributes.patron_count,
    Attributes.creation_count,
    Attributes.outstanding_payment_amount_cents,
]

default_relationships = [
    Relationships.rewards,
    Relationships.creator,
    Relationships.goals,
]



import requests

from patreon.utils import user_agent_string


class OAuth(object):
    def __init__(self, client_id, client_secret):
        super(OAuth, self).__init__()
        self.client_id = client_id
        self.client_secret = client_secret

    def get_tokens(self, code, redirect_uri):
        return self.__update_token({
            "grant_type": "authorization_code",
            "code": code,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "redirect_uri": redirect_uri
        })

    def refresh_token(self, refresh_token, redirect_uri=None):
        return self.__update_token({
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": self.client_id,
            "client_secret": self.client_secret
        })

    @staticmethod
    def __update_token(params):
        response = requests.post(
            "https://www.patreon.com/api/oauth2/token",
            params=params,
            headers={
                'User-Agent': user_agent_string(),
            }
        )
        return response.json()



from patreon.jsonapi.parser import JSONAPIParser, JSONAPIResource, JSONAPIRelationshipInfo, JSONAPIRelatedResource

jsonapi_doc = {
    "links": {
        "self": "http://example.com/articles",
        "next": "http://example.com/articles?page[offset]=2",
        "last": "http://example.com/articles?page[offset]=10"
    },
    "data": [{
        "type": "articles",
        "id": "1",
        "attributes": {
            "title": "JSON API paints my bikeshed!"
        },
        "relationships": {
            "author": {
                "links": {
                    "self": "http://example.com/articles/1/relationships/author",
                    "related": "http://example.com/articles/1/author"
                },
                "data": {"type": "people", "id": "9"}
            },
            "comments": {
                "links": {
                    "self": "http://example.com/articles/1/relationships/comments",
                    "related": "http://example.com/articles/1/comments"
                },
                "data": [
                    {"type": "comments", "id": "5"},
                    {"type": "comments", "id": "12"}
                ]
            }
        },
        "links": {
            "self": "http://example.com/articles/1"
        }
    }],
    "included": [{
        "type": "people",
        "id": "9",
        "attributes": {
            "first-name": "Dan",
            "last-name": "Gebhardt",
            "twitter": "dgeb"
        },
        "links": {
            "self": "http://example.com/people/9"
        }
    }, {
        "type": "comments",
        "id": "5",
        "attributes": {
            "body": "First!"
        },
        "relationships": {
            "author": {
                "data": {"type": "people", "id": "2"}
            }
        },
        "links": {
            "self": "http://example.com/comments/5"
        }
    }, {
        "type": "comments",
        "id": "12",
        "attributes": {
            "body": "I like XML better"
        },
        "relationships": {
            "author": {
                "data": {"type": "people", "id": "9"}
            }
        },
        "links": {
            "self": "http://example.com/comments/12"
        }
    }]
}
parsed = JSONAPIParser(jsonapi_doc)
article_1 = parsed.find_resource_by_type_and_id('articles', '1')


def test_document_has_accessible_raw_data():
    assert parsed.json_data is jsonapi_doc


def test_document_has_accessible_parsed_data():
    parsed_data = parsed.data()
    assert isinstance(parsed_data, list)
    assert len(parsed_data) == 1

    parsed_datum = parsed_data[0]
    assert isinstance(parsed_datum, JSONAPIResource)


def test_document_can_find_by_type_and_id():
    comments_12 = parsed.find_resource_by_type_and_id('comments', '12')
    assert isinstance(comments_12, JSONAPIResource)
    assert comments_12.type() == 'comments'
    assert comments_12.id() == '12'


def test_resource_type():
    assert article_1.type() == 'articles'


def test_resource_id():
    assert article_1.id() == '1'


def test_resource_attributes():
    assert article_1.attributes() == {
        'title': "JSON API paints my bikeshed!"
    }


def test_resource_attribute():
    assert article_1.attribute('title') == "JSON API paints my bikeshed!"


def test_resource_relationships():
    assert article_1.relationships() == {
        "author": {
            "links": {
                "self": "http://example.com/articles/1/relationships/author",
                "related": "http://example.com/articles/1/author"
            },
            "data": {"type": "people", "id": "9"}
        },
        "comments": {
            "links": {
                "self": "http://example.com/articles/1/relationships/comments",
                "related": "http://example.com/articles/1/comments"
            },
            "data": [
                {"type": "comments", "id": "5"},
                {"type": "comments", "id": "12"}
            ]
        }
    }


def test_resource_relationship():
    related_author = article_1.relationship('author')
    assert isinstance(related_author, JSONAPIResource)
    assert related_author.type() == 'people'
    assert related_author.id() == '9'
    assert related_author.attribute('first-name') == 'Dan'
    assert related_author.attribute('last-name') == 'Gebhardt'
    assert related_author.attribute('twitter') == 'dgeb'

    related_comments = article_1.relationship('comments')
    assert isinstance(related_comments, list)

    related_comment_5 = related_comments[0]
    assert isinstance(related_comment_5, JSONAPIResource)
    assert related_comment_5.type() == 'comments'
    assert related_comment_5.id() == '5'
    assert related_comment_5.attribute('body') == 'First!'

    related_comment_12 = related_comments[1]
    assert isinstance(related_comment_12, JSONAPIResource)
    assert related_comment_12.type() == 'comments'
    assert related_comment_12.id() == '12'
    assert related_comment_12.attribute('body') == 'I like XML better'


def test_resource_relationship_info():
    related_author_info = article_1.relationship_info('author')
    assert isinstance(related_author_info, JSONAPIRelationshipInfo)


def test_relationship_info_json_data():
    related_author_info = article_1.relationship_info('author')
    assert related_author_info.json_data == {
        "links": {
            "self": "http://example.com/articles/1/relationships/author",
            "related": "http://example.com/articles/1/author"
        },
        "data": {"type": "people", "id": "9"}
    }


def test_relationship_info_resource():
    related_author_info = article_1.relationship_info('author')
    related_author = related_author_info.resource()
    assert isinstance(related_author, JSONAPIResource)
    assert related_author.attribute('first-name') == 'Dan'
    assert related_author.attribute('last-name') == 'Gebhardt'
    assert related_author.attribute('twitter') == 'dgeb'


def test_relationship_info_resource_reference():
    related_author_info = article_1.relationship_info('author')
    related_author_reference = related_author_info.resource_reference()
    assert isinstance(related_author_reference, JSONAPIRelatedResource)
    assert related_author_reference.type() == 'people'
    assert related_author_reference.id() == '9'
    related_author = related_author_reference.resource()
    assert related_author.attribute('first-name') == 'Dan'
    assert related_author.attribute('last-name') == 'Gebhardt'
    assert related_author.attribute('twitter') == 'dgeb'



import pytest

from patreon.schemas import goal


@pytest.fixture
def attributes():
    return [
        'amount_cents',
        'title',
        'description',
        'created_at',
        'reached_at',
    ]


@pytest.fixture
def relationships():
    return []


def test_schema_attributes_are_properly_formatted(attributes):
    for attribute_name in attributes:
        value = getattr(goal.Attributes, attribute_name, None)
        assert value is not None and value is attribute_name


def test_schema_relationships_are_properly_formatted(relationships):
    assert len(relationships) is 0


def test_schema_has_expected_default_attributes():
    assert goal.Attributes.amount_cents == 'amount_cents'
    assert goal.Attributes.title == 'title'
    assert goal.Attributes.description == 'description'
    assert goal.Attributes.created_at == 'created_at'
    assert goal.Attributes.reached_at == 'reached_at'


def test_schema_has_expected_default_relationships():
    assert len(goal.default_relationships) is 0



from collections import OrderedDict

from patreon.version_compatibility.urllib_parse import urlencode


def joined_or_null(arr):
    return "null" if len(arr) == 0 else ','.join(arr)


def build_url(path, includes=None, fields=None):
    connector = '&' if '?' in path else '?'
    params = {}

    if includes:
        params.update({'include': joined_or_null(includes)})

    if fields:
        params.update(
            {
                "fields[{resource_type}]".format(resource_type=resource_type):
                joined_or_null(attributes)
                for resource_type, attributes in fields.items()
            }
        )

    if not params:
        return path

    sorted_params = OrderedDict(sorted(params.items(), key=lambda t: t[0]))
    return path + connector + urlencode(sorted_params)



from patreon.jsonapi import url_util

MOCK_URL_ENDPOINT = 'https://patreon.com'


def test_build_url_generates_url_without_params():
    assert url_util.build_url(MOCK_URL_ENDPOINT) == MOCK_URL_ENDPOINT


def test_build_url_generates_url_with_fields():
    expectation = MOCK_URL_ENDPOINT + '?fields%5Bmock%5D=test'

    assert expectation == url_util.build_url(
        MOCK_URL_ENDPOINT,
        fields={'mock': ['test']},
    )


def test_build_url_generates_url_with_includes():
    MOCK_INCLUDES = ['mock', 'include']
    expectation = MOCK_URL_ENDPOINT + '?include=' + '%2C'.join(MOCK_INCLUDES)

    assert expectation == url_util.build_url(
        MOCK_URL_ENDPOINT,
        includes=MOCK_INCLUDES,
    )


def test_build_url_generates_url_with_ampersand_if_query_params_present():
    MOCK_INCLUDES = ['example']
    expectation = MOCK_URL_ENDPOINT + '?mock=this&include=example'

    assert expectation == url_util.build_url(
        MOCK_URL_ENDPOINT + '?mock=this',
        includes=MOCK_INCLUDES,
    )



import sys

from setuptools import setup, find_packages

def is_running_tests():
    return {'pytest', 'test', 'ptr'}.intersection(sys.argv)

setup_requires = []

if is_running_tests:
    setup_requires.append('pytest-runner')

DESCRIPTION = (
    'Python library for interacting with the Patreon API. '
    'OAuth-centric for now.'
)

setup(
    name='patreon',
    version='0.5.0',
    description=DESCRIPTION,
    url='http://github.com/Patreon/patreon-python',
    author='Patreon',
    author_email='platform@patreon.com',
    license='Apache 2.0',
    packages=find_packages(
        exclude=['examples', 'examples.*', 'test', 'test.*']
    ),
    setup_requires=setup_requires,
    install_requires=[
        'requests',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
        'mock',
    ],
    zip_safe=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ]
)



class Attributes(object):
    email = 'email'
    first_name = 'first_name'
    last_name = 'last_name'
    full_name = 'full_name'
    gender = 'gender'
    status = 'status'
    vanity = 'vanity'
    about = 'about'
    facebook_id = 'facebook_id'
    image_url = 'image_url'
    thumb_url = 'thumb_url'
    thumbnails = 'thumbnails'
    youtube = 'youtube'
    twitter = 'twitter'
    facebook = 'facebook'
    twitch = 'twitch'
    is_suspended = 'is_suspended'
    is_deleted = 'is_deleted'
    is_nuked = 'is_nuked'
    created = 'created'
    url = 'url'
    like_count = 'like_count'
    comment_count = 'comment_count'
    is_creator = 'is_creator'
    hide_pledges = 'hide_pledges'
    two_factor_enabled = 'two_factor_enabled'


class Relationships(object):
    pledges = 'pledges'
    cards = 'cards'
    follows = 'follows'
    campaign = 'campaign'
    presence = 'presence'
    session = 'session'
    locations = 'locations'
    current_user_follow = 'current_user_follow'
    pledge_to_current_user = 'pledge_to_current_user'


default_attributes = [
    Attributes.email,
    Attributes.first_name,
    Attributes.last_name,
    Attributes.full_name,
    Attributes.gender,
    Attributes.status,
    Attributes.vanity,
    Attributes.about,
    Attributes.facebook_id,
    Attributes.image_url,
    Attributes.thumb_url,
    Attributes.thumbnails,
    Attributes.youtube,
    Attributes.twitter,
    Attributes.facebook,
    Attributes.twitch,
    Attributes.is_suspended,
    Attributes.is_deleted,
    Attributes.is_nuked,
    Attributes.created,
    Attributes.url,
]

default_relationships = [
    Relationships.campaign,
    Relationships.pledges,
]



import datetime


def has_timezone_support():
    return hasattr(datetime, 'timezone')


def utc_timezone():
    if has_timezone_support():
        return datetime.timezone.utc

    # Fallback to a mock Python 3 timezone for Python 2
    class UTCTimezone(datetime.tzinfo):
        def __init__(self):
            pass

        def utcoffset(self, dt):
            return datetime.timedelta(0)

        def dst(self, dt):
            return datetime.timedelta(0)

        def tzname(self, dt):
            return "UTC"

    return UTCTimezone()



class Attributes(object):
    amount_cents = 'amount_cents'
    total_historical_amount_cents = 'total_historical_amount_cents'
    declined_since = 'declined_since'
    created_at = 'created_at'
    pledge_cap_cents = 'pledge_cap_cents'
    patron_pays_fees = 'patron_pays_fees'
    unread_count = 'unread_count'


class Relationships(object):
    patron = 'patron'
    reward = 'reward'
    creator = 'creator'
    address = 'address'
    card = 'card'
    pledge_vat_location = 'pledge_vat_location'


default_attributes = [
    Attributes.amount_cents,
    Attributes.declined_since,
    Attributes.created_at,
    Attributes.pledge_cap_cents,
    Attributes.patron_pays_fees,
]

default_relationships = [
    Relationships.patron,
    Relationships.reward,
    Relationships.creator,
    Relationships.address,
    Relationships.pledge_vat_location,
]



import pytest

from patreon.schemas import pledge


@pytest.fixture
def attributes():
    return [
        'amount_cents',
        'total_historical_amount_cents',
        'declined_since',
        'created_at',
        'pledge_cap_cents',
        'patron_pays_fees',
        'unread_count',
    ]


@pytest.fixture
def relationships():
    return [
        'patron',
        'reward',
        'creator',
        'address',
        'card',
        'pledge_vat_location',
    ]


def test_schema_attributes_are_properly_formatted(attributes):
    for attribute_name in attributes:
        value = getattr(pledge.Attributes, attribute_name, None)
        assert value is not None and value is attribute_name


def test_schema_relationships_are_properly_formatted(relationships):
    for relationship_name in relationships:
        value = getattr(pledge.Relationships, relationship_name, None)
        assert value is not None and value is relationship_name


def test_schema_has_expected_default_attributes():
    assert pledge.Attributes.amount_cents in pledge.default_attributes
    assert pledge.Attributes.declined_since in pledge.default_attributes
    assert pledge.Attributes.created_at in pledge.default_attributes
    assert pledge.Attributes.pledge_cap_cents in pledge.default_attributes
    assert pledge.Attributes.patron_pays_fees in pledge.default_attributes


def test_schema_has_expected_default_relationships():
    assert pledge.Relationships.address in pledge.default_relationships
    assert pledge.Relationships.creator in pledge.default_relationships
    assert pledge.Relationships.patron in pledge.default_relationships
    assert pledge.Relationships.pledge_vat_location in pledge.default_relationships
    assert pledge.Relationships.reward in pledge.default_relationships



import pytest

from patreon.schemas import reward


@pytest.fixture
def attributes():
    return [
        'amount',
        'amount_cents',
        'user_limit',
        'remaining',
        'description',
        'requires_shipping',
        'created_at',
        'url',
        'patron_count',
    ]


@pytest.fixture
def relationships():
    return [
        'creator',
    ]


def test_schema_attributes_are_properly_formatted(attributes):
    for attribute_name in attributes:
        value = getattr(reward.Attributes, attribute_name, None)
        assert value is not None and value is attribute_name


def test_schema_relationships_are_properly_formatted(relationships):
    for relationship_name in relationships:
        value = getattr(reward.Relationships, relationship_name, None)
        assert value is not None and value is relationship_name


def test_schema_has_expected_default_attributes():
    assert reward.Attributes.amount in reward.default_attributes
    assert reward.Attributes.amount_cents in reward.default_attributes
    assert reward.Attributes.user_limit in reward.default_attributes
    assert reward.Attributes.remaining in reward.default_attributes
    assert reward.Attributes.description in reward.default_attributes
    assert reward.Attributes.requires_shipping in reward.default_attributes
    assert reward.Attributes.created_at in reward.default_attributes
    assert reward.Attributes.url in reward.default_attributes
    assert reward.Attributes.patron_count in reward.default_attributes


def test_schema_has_expected_default_relationships():
    assert reward.Relationships.creator in reward.default_relationships
